#include <windows.h>
#include "runtime.h"
#include "smaf.h"
#include "main.h"
#include "emusmw5.h"

// status
static int EmuPlayTempo = 100;
static int EmuPlayKeypitch = 0;
static int EmuPlayVolume = 127;

// Emulation
static HANDLE hEmuDLL;
static BYTE* EmuBuf;
static BYTE* EmuP;

static BOOL EmuSMAFPlaying;
static BOOL EmuMIDIPlaying;

// MIDI Emulation
static DWORD EmuMIDIStartTick;
static EVENT* EmuMIDIEvent;
static SMAF* EmuMIDICurrentSMAF;
static NOTEOFF* EmuMIDINoteoffs = NULL;
static int EmuMIDIVelocities[16];
static int EmuMIDIBankMSB[16];
static int EmuMIDISoloCh = -1;

// MaSound Prototypes
static int (*MaSound_EmuInitialize)(DWORD, DWORD, BYTE*);
static int (*MaSound_Initialize)(int, BYTE*, int);
static int (*MaSound_DeviceControl)(int, int, int, int);
static int (*MaSound_Terminate)();
static int (*MaSound_Create)(int);
static int (*MaSound_Load)(int, BYTE*, DWORD, int, int, int);
static int (*MaSound_Control)(int, int, int, int*, int);
static int (*MaSound_Open)(int, int, int, int);
static int (*MaSound_Standby)(int, int, int);
static int (*MaSound_Start)(int, int, int, int);
static int (*MaSound_GetEmuInfo)(int);
static int (*MaSound_Stop)(int, int, int);
static int (*MaSound_Seek)(int, int, int, int, int);
static int (*MaSound_Close)(int, int, int);
static int (*MaSound_Unload)(int, int, int);
static int (*MaSound_Delete)(int);
static int (*MaSound_Terminate)();
static int (*MaSound_EmuTerminate)();
static int (*SetMidiMsg)(BYTE*, DWORD);

// Functions
static BOOL InitEmuFunctions(HANDLE hdll)
{
	// Only need functions...

	if(!((FARPROC)MaSound_EmuInitialize	= GetProcAddress(hdll, "MaSound_EmuInitialize")))	return FALSE;
	if(!((FARPROC)MaSound_Initialize	= GetProcAddress(hdll, "MaSound_Initialize")))		return FALSE;
	if(!((FARPROC)MaSound_DeviceControl	= GetProcAddress(hdll, "MaSound_DeviceControl")))	return FALSE;
	if(!((FARPROC)MaSound_Terminate		= GetProcAddress(hdll, "MaSound_Terminate")))		return FALSE;
	if(!((FARPROC)MaSound_Create		= GetProcAddress(hdll, "MaSound_Create")))			return FALSE;
	if(!((FARPROC)MaSound_Load			= GetProcAddress(hdll, "MaSound_Load")))			return FALSE;
	if(!((FARPROC)MaSound_Control		= GetProcAddress(hdll, "MaSound_Control")))			return FALSE;
	if(!((FARPROC)MaSound_Open			= GetProcAddress(hdll, "MaSound_Open")))			return FALSE;
	if(!((FARPROC)MaSound_Standby		= GetProcAddress(hdll, "MaSound_Standby")))			return FALSE;
	if(!((FARPROC)MaSound_Start			= GetProcAddress(hdll, "MaSound_Start")))			return FALSE;
	if(!((FARPROC)MaSound_GetEmuInfo	= GetProcAddress(hdll, "MaSound_GetEmuInfo")))		return FALSE;
	if(!((FARPROC)MaSound_Stop			= GetProcAddress(hdll, "MaSound_Stop")))			return FALSE;
	if(!((FARPROC)MaSound_Seek			= GetProcAddress(hdll, "MaSound_Seek")))			return FALSE;
	if(!((FARPROC)MaSound_Close			= GetProcAddress(hdll, "MaSound_Close")))			return FALSE;
	if(!((FARPROC)MaSound_Unload		= GetProcAddress(hdll, "MaSound_Unload")))			return FALSE;
	if(!((FARPROC)MaSound_Delete		= GetProcAddress(hdll, "MaSound_Delete")))			return FALSE;
	if(!((FARPROC)MaSound_Terminate		= GetProcAddress(hdll, "MaSound_Terminate")))		return FALSE;
	if(!((FARPROC)MaSound_EmuTerminate	= GetProcAddress(hdll, "MaSound_EmuTerminate")))	return FALSE;
	if(!((FARPROC)SetMidiMsg			= GetProcAddress(hdll, "SetMidiMsg")))				return FALSE;

	return TRUE;
}

BOOL Emu526829(void)
{
	if(MaSound_DeviceControl(0x05, 2, 0, 0))
	{
		MessageBox(NULL, "DeviceControl(0x05, 2, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_DeviceControl(0x06, 0, 0, 0))
	{
		MessageBox(NULL, "DeviceControl(0x06, 0, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_DeviceControl(0x08, 2, 0, 0))
	{
		MessageBox(NULL, "DeviceControl(0x08, 2, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_DeviceControl(0x09, 0, 0, 0))
	{
		MessageBox(NULL, "DeviceControl(0x09, 0, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	return TRUE;
}

BOOL Emu5282(void)
{
	if(MaSound_DeviceControl(0x05, 2, 0, 0))
	{
		MessageBox(NULL, "DeviceControl(0x05, 2, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_DeviceControl(0x08, 2, 0, 0))
	{
		MessageBox(NULL, "DeviceControl(0x08, 2, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	return TRUE;
}

BOOL EmuInitialize(void)
{
	if(!(hEmuDLL = LoadLibrary("M5_EmuSmw5.dll")))
	{
		MessageBox(NULL, "M5_EmuSmw5.dll not found.\nIt requires to run this application.", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	if(!InitEmuFunctions(hEmuDLL))
		return FALSE;

	EmuBuf = halloc(1024);	// Ƃ肠m
	EmuP = EmuBuf;
	while(((DWORD)EmuP & 0xFF) != 0x81)	// MaSound_EmuInitializeParam3̓AhX1oCg0x81ȉ?ĂJust0x81łȂĂ͂ȂȂ炵BAlign?
		EmuP++;

	if(MaSound_EmuInitialize(48000, 0, EmuP))
	{
		MessageBox(NULL, "EmuInitialize0 failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	if(!Emu526829())
		return FALSE;

	// Init. FLAGS
	EmuSMAFPlaying = FALSE;
	EmuMIDIPlaying = FALSE;

	return TRUE;
}

void EmuUninitialize(void)
{
	MaSound_EmuTerminate();

	hfree(EmuBuf);

	FreeLibrary(hEmuDLL);
}

BOOL EmuSMAFPlay(SMAF* smaf, UINT startfrom)
{
	if(!smaf)
	{
		MessageBox(NULL, "EmuSMAFPlay: smaf == NULL, it is probably a program bug!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(!smaf->data)
	{
		MessageBox(NULL, "EmuSMAFPlay: No data. Please compile before EmuPlay.", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	MaSound_EmuTerminate();	// G[

	// Init
	if(MaSound_EmuInitialize(48000, 2, EmuP))
	{
		MessageBox(NULL, "EmuInitialize2 failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_Initialize(0, EmuBuf/* + 0x2F*/, 0))	// ͑v?
	{
		MessageBox(NULL, "Initialize failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_DeviceControl(0x0D, 0, 0, 0))	// ꂪPCMȉȂȂ悤ŁB
	{
		MessageBox(NULL, "DeviceControl(0x0D, 0, 0, 0) failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(!Emu526829())
	{
		return FALSE;
	}

	// open
	if(MaSound_Create(1) != 1)
	{
		MessageBox(NULL, "Create failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_Load(1, smaf->data, smaf->size, 1, (int)NULL, 0) != 1)	// ȂATS-MA5݂ƃR[obNۂǁAATS-MA5ł͈xĂ΂ĂȂBNULLw肵ĂOK̂ō̗pB
	{
		MessageBox(NULL, "Load failed!\nMay be incorrect SMAF specified.(CRC etc.)", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_Open(1, 1, 0, 0))
	{
		MessageBox(NULL, "Open failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_Standby(1, 1, 0))
	{
		MessageBox(NULL, "Standby failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	// Play
	if(MaSound_Control(1, 1, 0, &EmuPlayVolume, 0))	//  between 0 to 127
	{
		MessageBox(NULL, "Control volume failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_Control(1, 1, 2, &EmuPlayKeypitch, 0))	// L[sb` between -12 to 12
	{
		MessageBox(NULL, "Control keypitch failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(MaSound_Control(1, 1, 1, &EmuPlayTempo, 0))	// e|(%) between 70 to 130.
	{
		MessageBox(NULL, "Control tempo failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	if(MaSound_Seek(1, 1, startfrom, 0, 0))
	{
		MessageBox(NULL, "Seek failed", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	if(MaSound_Start(1, 1, 1, 0))
	{
		MessageBox(NULL, "Start failed!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	EmuSMAFPlaying = TRUE;

	return TRUE;
}
void EmuSMAFStop(void)
{
	MaSound_Stop(1, 1, 0);

	MaSound_Close(1, 1, 0);
	MaSound_Unload(1, 1, 0);
	MaSound_Delete(1);

	MaSound_Terminate();
	MaSound_EmuTerminate();

	EmuSMAFPlaying = FALSE;

	if(MaSound_EmuInitialize(48000, 0, EmuP))
	{
		MessageBox(NULL, "EmuInitialize0 failed!", "Error", MB_ICONERROR | MB_OK);
		return;
	}

	if(!Emu526829())
		return;

	return;
}

//SMAF/MIDIp
int EmuStatus(void)
{
	if(EmuSMAFPlaying)
		return MaSound_Control(1, 1, 6, 0, 0);
	if(EmuMIDIPlaying)
	{
		if(EmuMIDIEvent)
			return EMUSTATUS_PLAYING;
		else
			return EMUSTATUS_READY;
	}

	return -1;
}

int EmuInfo(void)
{
	return MaSound_GetEmuInfo(4);
}

int EmuSetMidiMsg(BYTE* data, DWORD size)
{
	return SetMidiMsg(data, size);
}

BOOL EmuSetVolume(int vol)
{
	EmuPlayVolume = vol;

	if(EmuSMAFPlaying && MaSound_Control(1, 1, 0, &vol, 0))
		return TRUE;
	if(EmuMIDIPlaying)
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x05, 0x7E, 0x09, EmuPlayVolume, 0xF7};
		if(SetMidiMsg(cmd, sizeof(cmd)))
			return TRUE;
	}

	return FALSE;
}
BOOL EmuSetTempo(int tempo)
{
	EmuPlayTempo = tempo;

	if(EmuSMAFPlaying && MaSound_Control(1, 1, 1, &tempo, 0))
		return TRUE;

	return FALSE;
}
BOOL EmuSetKeypitch(int key)
{
	EmuPlayKeypitch = key;

	if(EmuSMAFPlaying && MaSound_Control(1, 1, 2, &key, 0))
		return TRUE;

	return FALSE;
}

int EmuGetVolume(void)
{
	return EmuPlayVolume;
}
int EmuGetTempo(void)
{
	return EmuPlayTempo;
}
int EmuGetKeypitch(void)
{
	return EmuPlayKeypitch;
}

BOOL EmuMIDISetVolume(void)
{
	BYTE cmd[] = {0xF0, 0x43, 0x79, 0x05, 0x7E, 0x09, EmuPlayVolume, 0xF7};
	if(EmuSetMidiMsg(cmd, sizeof(cmd)))
	{
		MessageBox(NULL, "EmuMIDISetVolume failed", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	return TRUE;
}
BOOL DoEvent(EVENT* ev, BOOL issound)
{
	// sounding
	if((ev->status & 0xF0) == 0x80 || (ev->status & 0xF0) == 0x90)
	{
		int ch = ev->status & 0x0F;

		if((ev->status & 0xF0) == 0x90)
			EmuMIDIVelocities[ch] = ev->data[1];

		if(issound && (EmuMIDISoloCh == -1 || EmuMIDISoloCh == (ev->status & 0x0F)))
		{
			BYTE cmd[3];

			cmd[0] = 0x90 | ch;
			cmd[1] = ev->data[0];
			if(EmuMIDIBankMSB[ch] != 125)	// Â0x000x80ƂgĂ邩玀˂B
				cmd[1] += EmuPlayKeypitch;

			cmd[2] = EmuMIDIVelocities[ch];

			if(SetMidiMsg(cmd, 3))
			{
				MessageBox(NULL, "EmuMIDI: DoEvent NoteON Failed", "Error", MB_ICONERROR | MB_OK);
				return FALSE;
			}

			{
				NOTEOFF* no = EmuMIDINoteoffs;

				if(EmuMIDINoteoffs)
				{
					while(no->next)
						no = no->next;
					no->next = halloc(sizeof(NOTEOFF));
					no = no->next;
				}else
				{
					EmuMIDINoteoffs = halloc(sizeof(NOTEOFF));
					no = EmuMIDINoteoffs;
				}

				no->ch = ch;
				no->sound = ev->data[0] + EmuPlayKeypitch;
				no->time = ev->time + GetVVal(ev->data + 1 + (((ev->status & 0xF0) == 0x90) ? 1 : 0))/* * EmuMIDICurrentSMAF->gbase*/;
				no->next = NULL;
			}
		}
	}
	if((ev->status & 0xF0) == 0xB0)
	{
		BYTE cmd[3];
		int ch = ev->status & 0x0F;

		cmd[0] = ev->status;
		cmd[1] = ev->data[0];
		cmd[2] = ev->data[1];

		if(SetMidiMsg(cmd, 3))
		{
			MessageBox(NULL, "EmuMIDI: DoEvent CC Failed", "Error", MB_ICONERROR | MB_OK);
			return FALSE;
		}

		if(ev->data[0] == 0x00)
			EmuMIDIBankMSB[ch] = ev->data[1];
	}
	// 0xEn = PitchBend(SMAF only, not GM.)
	if((ev->status & 0xF0) == 0xE0)
	{
		BYTE cmd[3];
		int ch = ev->status & 0x0F;

		cmd[0] = ev->status;
		cmd[1] = ev->data[0];
		cmd[2] = ev->data[1];

		if(SetMidiMsg(cmd, 3))
		{
			MessageBox(NULL, "EmuMIDI: DoEvent PB Failed", "Error", MB_ICONERROR | MB_OK);
			return FALSE;
		}
	}
	if((ev->status & 0xF0) == 0xC0)
	{
		BYTE cmd[2];
		int ch = ev->status & 0x0F;

		cmd[0] = ev->status;
		cmd[1] = ev->data[0];

		if(SetMidiMsg(cmd, 2))
		{
			MessageBox(NULL, "EmuMIDI: DoEvent PC Failed", "Error", MB_ICONERROR | MB_OK);
			return FALSE;
		}
	}
	if(ev->status == 0xF0)
	{
		// KH code, it should check with 43 79 xx xx xx.
		if(GetVVal(ev->data) == 0x06	// Reset
			// || GetVVal(ev->data) == 0x07	// Volume
			|| GetVVal(ev->data) == 0x1E	// PCM define
			|| GetVVal(ev->data) == 0x2F	// FM define
			// Errorł܂
			//|| GetVVal(ev->data) == 0x1B || GetVVal(ev->data) == 0x2A
			// Ȃɂ
			//|| GetVVal(ev->data) == 0x1F
			// MA-2's FM
			|| GetVVal(ev->data) == 0x12
			// WT
			|| GetVVal(ev->data) >= 256
			)
		{
			// exclusiveȂ̂...resetł͂Ȃ?
			if(GetVVal(ev->data) == 0x1E || GetVVal(ev->data) == 0x2F)
			{
				BYTE cmd[0x0B] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x02, ev->data[6], ev->data[7], ev->data[8], ev->data[9], 0xF7};

				if(EmuSetMidiMsg(cmd, 0x0B))
				{
					TCHAR ts[256];

					wsprintf(ts, "EmuMIDI: DoEvent EmuSetMidiMsg(Erase?) failed");
					MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);
					return FALSE;
				}
			}

			{
				BYTE* cmd;
				DWORD size = GetVVal(ev->data);
				cmd = halloc(size + 1);
				memcopy(cmd, ev->data + (GetVValSize(ev->data) - 1), size + 1);
				cmd[0] = 0xF0;
				//cmd[3] = 0x07;

				if(SetMidiMsg(cmd, GetVVal(ev->data) + 1))
				{
					TCHAR ts[256];

					wsprintf(ts, "EmuMIDI: DoEvent Exclusive failed(size=%d)", GetVVal(ev->data) + 1);
					MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);
					hfree(cmd);
					return FALSE;
				}
				hfree(cmd);
			}

			// Reset/SetMasterVolume...oh reset volume!!
			if(GetVVal(ev->data) == 0x06 && ev->data[4] == 0x7F && ev->data[5] == 0x7F)
				EmuMIDISetVolume();
		}
	}

	return TRUE;
}
static BOOL DoMIDIEvent(BOOL issound)
{
	if(!DoEvent(EmuMIDIEvent, issound))
		return FALSE;

	EmuMIDIEvent = EmuMIDIEvent->next;

	return TRUE;
}
BOOL EmuMIDIPlay(SMAF* smaf, UINT startfrom)
{
	EmuMIDICurrentSMAF = smaf;
	EmuMIDIEvent = smaf->events;

	if(!smaf)
	{
		MessageBox(NULL, "EmuMIDIPlay: smaf == NULL, it is probably a program bug!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(!smaf->events)
	{
		MessageBox(NULL, "EmuMIDIPlay: No events. Please put one or more notes onto pianoroll.", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}
	if(!smaf->start)
	{
		MessageBox(NULL, "EmuMIDIPlay: smaf->events != NULL but smaf->start == NULL, it is probably a program bug!", "Error", MB_ICONERROR | MB_OK);
		return FALSE;
	}

	// Velocity Default
	{
		int i;

		for(i = 0; i < 16; i++)
		{
			EmuMIDIVelocities[i] = 64;
			EmuMIDIBankMSB[i] = 124;
		}
		EmuMIDIBankMSB[9] = 125;	// Tr10=Drum default.
	}

	// Reset
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x7F, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(NULL, "SetMidiMsg ini failed", "Error", MB_ICONERROR | MB_OK);
			return FALSE;
		}
	}

	// forwarding event
	while(EmuMIDIEvent && EmuMIDIEvent->time < startfrom + (smaf->start->time - smaf->start->duration))
	{
		if(!DoMIDIEvent(FALSE))
		{
			MessageBox(NULL, "EmuMIDI: EmuMIDIPlay DoEvent Failed", "Error", MB_ICONERROR | MB_OK);
			return FALSE;
		}
	}

	EmuMIDINoteoffs = NULL;

	// set volume
	EmuMIDISetVolume();

	// start!
	EmuMIDIStartTick = GetTickCount() - ((smaf->start->time - smaf->start->duration) + startfrom);
	EmuMIDIPlaying = TRUE;

	return TRUE;
}

static BOOL EmuDeleteAllNoteoff(void)
{
	NOTEOFF* no = EmuMIDINoteoffs;

	while(no)
	{
		NOTEOFF* nx = no->next;

		{
			BYTE cmd[3];
			cmd[0] = 0x90 | no->ch;
			cmd[1] = no->sound;
			cmd[2] = 0;

			if(SetMidiMsg(cmd, 3))
			{
				MessageBox(NULL, "EmuMIDI: EmuDeleteAllNoteoff NoteOFF Failed", "Error", MB_ICONERROR | MB_OK);
				return FALSE;
			}
		}

		hfree(no);
		no = nx;
	}

	EmuMIDINoteoffs = NULL;

	return TRUE;
}

BOOL EmuMIDIStop(void)
{
	EmuMIDIPlaying = FALSE;

	if(!EmuDeleteAllNoteoff())
		return FALSE;

	return TRUE;
}

//SMAF/MIDIp
BOOL EmuSeek(UINT pos)
{
	if(EmuSMAFPlaying)
	{
		MaSound_Stop(1, 1, 0);
		if(MaSound_Seek(1, 1, pos, 0, 0))
			return FALSE;
		MaSound_Start(1, 1, 1, 0);

		return TRUE;
	}
	if(EmuMIDIPlaying)
	{
		SMAF* smaf = EmuMIDICurrentSMAF;

		EmuMIDIStop();

		return EmuMIDIPlay(smaf, pos);
	}

	return FALSE;
}

//SMAF/MIDIp
int EmuGetCurrentPos(void)
{
	if(EmuSMAFPlaying)
		return MaSound_Control(1, 1, 4, 0, 0);
	if(EmuMIDIPlaying)
		return GetTickCount() - EmuMIDIStartTick;

	return -1;
}

BOOL EmuMIDIProc(void)
{
	DWORD curpos = GetTickCount() - EmuMIDIStartTick;

	{
		NOTEOFF* no = EmuMIDINoteoffs;
		NOTEOFF* pre = NULL;

		while(no)
		{
			NOTEOFF* nx = no->next;
			if(no->time < curpos)
			{
				BYTE cmd[3];
				cmd[0] = 0x90 | no->ch;
				cmd[1] = no->sound;
				cmd[2] = 0;

				if(SetMidiMsg(cmd, 3))
				{
					MessageBox(NULL, "EmuMIDI: EmuMIDIProc NoteOFF Failed", "Error", MB_ICONERROR | MB_OK);
					return FALSE;
				}

				hfree(no);

				if(!pre)
					EmuMIDINoteoffs = nx;
				else
					pre->next = nx;
			}else
			{
				pre = no;
			}
			no = nx;
		}
	}

	while(EmuMIDIEvent && EmuMIDIEvent->time < curpos)
	{
		if(!DoMIDIEvent(TRUE))
		{
			MessageBox(NULL, "EmuMIDI: EmuMIDIEvent DoEvent Failed", "Error", MB_ICONERROR | MB_OK);
			return FALSE;
		}
	}

	return TRUE;
}

void EmuMIDISetSoloCh(int ch)
{
	EmuMIDISoloCh = ch;

	return;
}
